/* ------------------------------------- **
**      Prototype for opt.*.c files      **
** ------------------------------------- */

/* -------------- **
**  opt.memory.c  **
** -------------- */

long global_data_size(void);
long local_data_size(void);
void open_mapfile(void);
void close_mapfile(void);
void map_memory(long HeapArea, long GlobalLocalArea, long TrailAuxArea, int n_workers);
void unmap_memory(void);
void alloc_all_optyap_structs(void);
#ifdef YAPOR
void remap_memory(void);
void alloc_or_frames(void);
void alloc_answer_frames(void);
void alloc_solution_frames(void);
#endif /* YAPOR */
#ifdef TABLING
void alloc_table_entries(void);
void alloc_subgoal_trie_nodes(void);
void alloc_subgoal_frames(void);
void alloc_answer_trie_nodes(void);
void alloc_dependency_frames(void);
#ifdef YAPOR
void alloc_suspension_frames(void);
void *alloc_suspension_block(int size);
void reset_suspension_area(void);
#endif /* YAPOR */
#endif /* TABLING */


/* ------------ **
**  opt.misc.c  **
** ------------ */

void abort_optyap(const char *msg, ...);
void itos(int i, char *s);
void information_message(const char *mesg,...);
#if defined(YAPOR_ERRORS) || defined(TABLING_ERRORS)
void error_message(const char *mesg, ...);
#endif /* YAPOR_ERRORS || TABLING_ERRORS */
#if defined(YAPOR_DEBUG) || defined(TABLING_DEBUG)
void debug_message(const char *mesg, ...);
#endif /* YAPOR_DEBUG || TABLING_DEBUG */
#ifdef TABLING_DEBUG
void tabling_instr_debug_message(char *inst, void *pred, Term *args);
#endif /* TABLING_DEBUG */


/* ------------ **
**  opt.init.c  **
** ------------ */

void init_global(int n_workers, int sch_loop, int delay_load);
void init_local(void);
void make_root_frames(void);
#ifdef YAPOR
void init_workers(void);
#endif /* YAPOR */
void init_signals(void);


/* ------------- **
**  opt.preds.c  **
** ------------- */

void init_optyap_preds(void);
#ifdef YAPOR
int yapor_on(void);
void yapor_off(void);
#endif /* YAPOR */



/* ------------------------------------- **
**      Prototype for tab.*.c files      **
** ------------------------------------- */

/* ------------- **
**  tab.tries.c  **
** ------------- */

#ifdef TABLING
sg_node_ptr subgoal_search(OPREG arity, sg_node_ptr current_sg_node, CELL **Yaddr);
ans_node_ptr answer_search(ans_node_ptr current_ans_node, CELL *subs_ptr);
int p_abolish_incomplete_tables(void);
int p_check_incomplete_tables(void);
int p_abolish_trie(void);
void show_trie(sg_node_ptr sg_node, int pred_arity, Atom pred_atom);
void show_subgoal_trie(sg_node_ptr sg_node, char *str, int str_index, int *arity);
void show_answer_trie(ans_node_ptr ans_node, char *str, int str_index, int *arity, int var_index);
#endif /* TABLING */


/* --------------- **
**  tab.suspend.c  **
** --------------- */

#ifdef TABLING
void private_completion(void);
#ifdef YAPOR
void public_completion(void);
void complete_suspension_frames(susp_fr_ptr chain_susp_fr);
void suspend_completion_branch(void);
void resume_suspension_frame(susp_fr_ptr resume_fr);
void collect_suspension_frames_with_unconsumed_answers(void);
int suspension_frame_with_unconsumed_answers(susp_fr_ptr susp_fr);
#endif /* YAPOR */
#endif /* TABLING */



/* ------------------------------------ **
**      Prototype for or.*.c files      **
** ------------------------------------ */

/* ------------- **
**  or.engine.c  **
** ------------- */

#ifdef ENV_COPY
void make_root_choice_point(void);
void free_root_choice_point(void);
int q_share_work(int p);
void p_share_work(void);
#endif /* ENV_COPY */


/* ---------------- **
**  or.cowengine.c  **
** ---------------- */

#ifdef ACOW
void make_root_choice_point(void);
void free_root_choice_point(void);
int q_share_work(int p);
int p_share_work(void);
#endif /* ACOW */


/* ---------------- **
**  or.sbaengine.c  **
** ---------------- */

#ifdef SBA
void make_root_choice_point(void);
void free_root_choice_point(void);
int q_share_work(int p);
void p_share_work(void);
#endif /* SBA */


/* ---------------- **
**  or.scheduler.c  **
** ---------------- */

#ifdef YAPOR
int move_up_one_node(or_fr_ptr nearest_livenode);
int get_work(void);
#endif /* YAPOR */


/* ---------- **
**  or.cut.c  **
** ---------- */

#ifdef YAPOR
int no_workers_left_pending_cut(void);
void cut_shared(choiceptr cut_cp, sol_fr_ptr aux_solution);
#endif /* YAPOR */
