/* --------------------------- **
**      Execution defines      **
** --------------------------- */

#define YAPOR_STATISTICS
#define TABLING_STATISTICS
#define YAPOR_ERRORS
#define TABLING_ERRORS
/*
#define YAPOR_DEBUG
#define TABLING_DEBUG
*/



#ifndef YAPOR
#undef YAPOR_STATISTICS
#undef YAPOR_ERRORS
#undef YAPOR_DEBUG
#endif /* YAPOR */

#ifndef TABLING
#undef TABLING_STATISTICS
#undef TABLING_ERRORS
#undef TABLING_DEBUG
#endif /* TABLING */

#if defined(YAPOR_STATISTICS) && defined(TABLING_STATISTICS)
#define OPTYAP_STATISTICS
#endif /* YAPOR_STATISTICS && TABLING_STATISTICS */

#if defined(YAPOR_ERRORS) && defined(TABLING_ERRORS)
#define OPTYAP_ERRORS
#endif /* YAPOR_ERRORS && TABLING_ERRORS */

#if defined(YAPOR_DEBUG) && defined(TABLING_DEBUG)
#define OPTYAP_DEBUG
#endif /* YAPOR_DEBUG && TABLING_DEBUG */



/* ------------------------------ **
**      Typedefs and defines      **
** ------------------------------ */

typedef volatile int lockvar;
typedef unsigned long bitmap;
typedef double realtime;

#define MAX_WORKERS (sizeof(bitmap) * 8 - 1)
#define MAX_LENGTH_ANSWER 500
#define MAX_DEPTH 500
#define MAX_BEST_TIMES 21
#define MAX_TABLE_VARS 512



/* --------------------------------- **
**      Struct global_statistic      **
** --------------------------------- */

#if defined(YAPOR_STATISTICS) || defined(TABLING_STATISTICS)
struct global_statistics {
  lockvar lock;
#ifdef YAPOR_STATISTICS
  int allocated_or_frames;
  int used_or_frames;
  int allocated_answer_frames;
  int used_answer_frames;
  int allocated_solution_frames;
  int used_solution_frames;
#endif /* YAPOR_STATISTICS */
#ifdef TABLING_STATISTICS
  int allocated_table_entries;
  int used_table_entries;
  int allocated_subgoal_trie_nodes;
  int used_subgoal_trie_nodes;
  int allocated_subgoal_frames;
  int used_subgoal_frames;
  int allocated_answer_trie_nodes;
  int used_answer_trie_nodes;
  int allocated_dependency_frames;
  int used_dependency_frames;
#endif /* TABLING_STATISTICS */
#ifdef OPTYAP_STATISTICS
  int allocated_suspension_frames;
  int used_suspension_frames;
#endif /* OPTYAP_STATISTICS */
};
#endif /* YAPOR_STATISTICS || TABLING_STATISTICS */



/* ----------------------------- **
**      Struct global_locks      **
** ----------------------------- */

#ifdef YAPOR
struct global_locks {
#if defined(YAPOR_ERRORS) || defined(TABLING_ERRORS) || defined(YAPOR_DEBUG) || defined(TABLING_DEBUG)
  lockvar stderr_messages;
#endif /* YAPOR_ERRORS || TABLING_ERRORS || YAPOR_DEBUG || TABLING_DEBUG */

  lockvar who_locked_heap;  /* used in CriticalSection's macros */
  lockvar heap_access;      /* used in CriticalSection's macros */

  lockvar bitmap_idle_workers;
  lockvar bitmap_root_cp_workers;
  lockvar bitmap_invisible_workers;
  lockvar bitmap_requestable_workers;
  lockvar bitmap_executing_workers;
  lockvar bitmap_finished_workers;

  lockvar alloc_area;
  lockvar next_free_or_frame;
  lockvar next_free_answer_frame;
  lockvar next_free_solution_frame;
#ifdef TABLING
  lockvar next_free_table_entry;
  lockvar next_free_subgoal_trie_node;
  lockvar next_free_subgoal_frame;
  lockvar next_free_answer_trie_node;
  lockvar next_free_dependency_frame;
  lockvar next_free_suspension_frame;
#endif /* TABLING */
};
#endif /* YAPOR */



/* ---------------------------- **
**      Struct global_data      **
** ---------------------------- */

struct global_data{
  /* global data related to statistics */
#if defined(YAPOR_STATISTICS) || defined(TABLING_STATISTICS)
  struct global_statistics statistics;
#endif /* YAPOR_STATISTICS || TABLING_STATISTICS */

  /* global data related to memory management */
  void *BaseWorkArea;
  void *TopWorkArea;
  void *TopAllocArea;
#if defined(YAPOR) && defined(TABLING)
  void *SuspensionArea;
#endif /* YAPOR && TABLING */
#ifdef YAPOR
  struct global_locks locks;
  struct or_frame *next_free_or_frame;
  struct answer_frame *next_free_answer_frame;
  struct solution_frame *next_free_solution_frame;
#endif /* YAPOR */
#ifdef TABLING
  struct table_entry *next_free_table_entry;
  struct subgoal_trie_node *next_free_subgoal_trie_node;
  struct subgoal_frame *next_free_subgoal_frame;
  struct answer_trie_node *next_free_answer_trie_node;
  struct dependency_frame *next_free_dependency_frame;
#ifdef YAPOR
  struct suspension_frame *next_free_suspension_frame;
#endif /* YAPOR */
#endif /* TABLING */

#ifdef YAPOR
  /* global static data */
  int scheduler_loop;
  int delayed_release_load;
  int number_workers;
  int worker_pid[MAX_WORKERS];
#ifdef ACOW
  int master_worker;
#endif /* ACOW */

  /* global data related to or-performance */
  realtime execution_time;
  realtime best_execution_times[MAX_BEST_TIMES];
  int number_of_executed_goals;
  char performance_mode;  /* PERFORMANCE_OFF / PERFORMANCE_ON / PERFORMANCE_IN_EXECUTION */

  /* global data related to or-parallelism */
  choiceptr root_choice_point;
  struct or_frame *root_or_frame;
  volatile struct {
    int alt_ltt;
    int alt_cut;
  } branch[MAX_WORKERS][MAX_DEPTH];
  bitmap present_workers;
  volatile bitmap idle_workers;
  volatile bitmap root_cp_workers;
  volatile bitmap invisible_workers;
  volatile bitmap requestable_workers;
  volatile bitmap executing_workers;
  volatile bitmap finished_workers;
  volatile char execution_mode;  /* SEQUENTIAL_MODE / PARALLEL_MODE */
  volatile int answers;
#endif /* YAPOR */

#ifdef TABLING
  /* global data related to tabling */
  struct dependency_frame *root_dependency_frame;
  CELL table_var_enumerator[MAX_TABLE_VARS];
#endif /* TABLING */
};

extern struct global_data *GLOBAL;

#define GLOBAL_STATS_lock                     (GLOBAL->statistics.lock)
#define GLOBAL_STATS_allocated_or_fr          (GLOBAL->statistics.allocated_or_frames)
#define GLOBAL_STATS_used_or_fr               (GLOBAL->statistics.used_or_frames)
#define GLOBAL_STATS_allocated_ans_fr         (GLOBAL->statistics.allocated_answer_frames)
#define GLOBAL_STATS_used_ans_fr              (GLOBAL->statistics.used_answer_frames)
#define GLOBAL_STATS_allocated_sol_fr         (GLOBAL->statistics.allocated_solution_frames)
#define GLOBAL_STATS_used_sol_fr              (GLOBAL->statistics.used_solution_frames)
#define GLOBAL_STATS_allocated_tab_ent        (GLOBAL->statistics.allocated_table_entries)
#define GLOBAL_STATS_used_tab_ent             (GLOBAL->statistics.used_table_entries)
#define GLOBAL_STATS_allocated_sg_node        (GLOBAL->statistics.allocated_subgoal_trie_nodes)
#define GLOBAL_STATS_used_sg_node             (GLOBAL->statistics.used_subgoal_trie_nodes)
#define GLOBAL_STATS_allocated_sg_fr          (GLOBAL->statistics.allocated_subgoal_frames)
#define GLOBAL_STATS_used_sg_fr               (GLOBAL->statistics.used_subgoal_frames)
#define GLOBAL_STATS_allocated_ans_node       (GLOBAL->statistics.allocated_answer_trie_nodes)
#define GLOBAL_STATS_used_ans_node            (GLOBAL->statistics.used_answer_trie_nodes)
#define GLOBAL_STATS_allocated_dep_fr         (GLOBAL->statistics.allocated_dependency_frames)
#define GLOBAL_STATS_used_dep_fr              (GLOBAL->statistics.used_dependency_frames)
#define GLOBAL_STATS_allocated_susp_fr        (GLOBAL->statistics.allocated_suspension_frames)
#define GLOBAL_STATS_used_susp_fr             (GLOBAL->statistics.used_suspension_frames)
#define BaseWorkArea                          (GLOBAL->BaseWorkArea)
#define TopWorkArea                           (GLOBAL->TopWorkArea)
#define TopAllocArea                          (GLOBAL->TopAllocArea)
#define SuspensionArea                        (GLOBAL->SuspensionArea)
#define GLOBAL_LOCKS_stderr_messages          (GLOBAL->locks.stderr_messages)
#define GLOBAL_LOCKS_who_locked_heap          (GLOBAL->locks.who_locked_heap)
#define GLOBAL_LOCKS_heap_access              (GLOBAL->locks.heap_access)
#define GLOBAL_LOCKS_bm_idle_workers          (GLOBAL->locks.bitmap_idle_workers)
#define GLOBAL_LOCKS_bm_root_cp_workers       (GLOBAL->locks.bitmap_root_cp_workers)
#define GLOBAL_LOCKS_bm_invisible_workers     (GLOBAL->locks.bitmap_invisible_workers)
#define GLOBAL_LOCKS_bm_requestable_workers   (GLOBAL->locks.bitmap_requestable_workers)
#define GLOBAL_LOCKS_bm_executing_workers     (GLOBAL->locks.bitmap_executing_workers)
#define GLOBAL_LOCKS_bm_finished_workers      (GLOBAL->locks.bitmap_finished_workers)
#define GLOBAL_LOCKS_alloc_area               (GLOBAL->locks.alloc_area)
#define GLOBAL_LOCKS_next_free_or_fr          (GLOBAL->locks.next_free_or_frame)
#define GLOBAL_LOCKS_next_free_ans_fr         (GLOBAL->locks.next_free_answer_frame)
#define GLOBAL_LOCKS_next_free_sol_fr         (GLOBAL->locks.next_free_solution_frame)
#define GLOBAL_LOCKS_next_free_tab_ent        (GLOBAL->locks.next_free_table_entry)
#define GLOBAL_LOCKS_next_free_sg_node        (GLOBAL->locks.next_free_subgoal_trie_node)
#define GLOBAL_LOCKS_next_free_sg_fr          (GLOBAL->locks.next_free_subgoal_frame)
#define GLOBAL_LOCKS_next_free_ans_node       (GLOBAL->locks.next_free_answer_trie_node)
#define GLOBAL_LOCKS_next_free_dep_fr         (GLOBAL->locks.next_free_dependency_frame)
#define GLOBAL_LOCKS_next_free_susp_fr        (GLOBAL->locks.next_free_suspension_frame)
#define GLOBAL_next_free_or_fr                (GLOBAL->next_free_or_frame)
#define GLOBAL_next_free_ans_fr               (GLOBAL->next_free_answer_frame)
#define GLOBAL_next_free_sol_fr               (GLOBAL->next_free_solution_frame)
#define GLOBAL_next_free_tab_ent              (GLOBAL->next_free_table_entry)
#define GLOBAL_next_free_sg_node              (GLOBAL->next_free_subgoal_trie_node)
#define GLOBAL_next_free_sg_fr                (GLOBAL->next_free_subgoal_frame)
#define GLOBAL_next_free_ans_node             (GLOBAL->next_free_answer_trie_node)
#define GLOBAL_next_free_dep_fr               (GLOBAL->next_free_dependency_frame)
#define GLOBAL_next_free_susp_fr              (GLOBAL->next_free_suspension_frame)
#define SCHEDULER_LOOP                        (GLOBAL->scheduler_loop)
#define DELAYED_RELEASE_LOAD                  (GLOBAL->delayed_release_load)
#define number_workers                        (GLOBAL->number_workers)
#define worker_pid(w)                         (GLOBAL->worker_pid[w])
#define GLOBAL_master_worker                  (GLOBAL->master_worker)
#define GLOBAL_execution_time                 (GLOBAL->execution_time)
#define GLOBAL_best_times(t)                  (GLOBAL->best_execution_times[t])
#define GLOBAL_number_goals                   (GLOBAL->number_of_executed_goals)
#define GLOBAL_performance_mode               (GLOBAL->performance_mode)
#define GLOBAL_root_cp                        (GLOBAL->root_choice_point)
#define GLOBAL_root_or_fr                     (GLOBAL->root_or_frame)
#define GLOBAL_branch_ltt(w,d)                (GLOBAL->branch[w][d].alt_ltt)
#define GLOBAL_branch_cut(w,d)                (GLOBAL->branch[w][d].alt_cut)
#define GLOBAL_bm_present_workers             (GLOBAL->present_workers)
#define GLOBAL_bm_idle_workers                (GLOBAL->idle_workers)
#define GLOBAL_bm_root_cp_workers             (GLOBAL->root_cp_workers)
#define GLOBAL_bm_invisible_workers           (GLOBAL->invisible_workers)
#define GLOBAL_bm_requestable_workers         (GLOBAL->requestable_workers)
#define GLOBAL_bm_executing_workers           (GLOBAL->executing_workers)
#define GLOBAL_bm_finished_workers            (GLOBAL->finished_workers)
#define GLOBAL_execution_mode                 (GLOBAL->execution_mode)
#define GLOBAL_answers                        (GLOBAL->answers)
#define GLOBAL_root_dep_fr                    (GLOBAL->root_dependency_frame)
#define GLOBAL_table_var_enumerator(i)        (GLOBAL->table_var_enumerator[i])


/* ------------------------------ **
**      Struct local_signals      **
** ------------------------------ */

#ifdef YAPOR
struct local_signals{
  volatile enum {
    no_reply = 0,
    accepted = 1,
    local_ok = 2,
    install  = 3,
    ready    = 4,
    refused  = 5 
  } reply;
  volatile enum {
    Q_idle = 0,
    trail  = 1,
    global = 2,
    local  = 3,
    P_idle = 4
  } P_fase, Q_fase;
  lockvar lock;
};
#endif /* YAPOR */



/* --------------------------- **
**      Struct local_data      **
** --------------------------- */

struct local_data{
  lockvar lock;

#ifdef YAPOR
  /* local data related to or-parallelism */
  choiceptr top_choice_point;
  struct or_frame *top_or_frame;
  volatile int load;
  volatile int request_from_worker;
  struct local_signals sharing_signals;
  volatile struct {
    CELL start;
    CELL end;
  } global_copy, local_copy, trail_copy;
  choiceptr cut_choice_point;
#endif /* YAPOR */

#ifdef TABLING
  /* local data related to tabling */
#ifdef YAPOR
  choiceptr top_choice_point_on_stack;
  struct suspension_frame *top_collected_suspension_frame;
#endif /* YAPOR */
  struct dependency_frame *top_dependency_frame;
  struct subgoal_frame *top_subgoal_frame;
#endif /* TABLING */
};

extern struct local_data *LOCAL;
#ifdef YAPOR
extern struct local_data *REMOTE[MAX_WORKERS];
#endif /* YAPOR */

#define LOCAL_lock                       (LOCAL->lock)
#define LOCAL_top_cp                     (LOCAL->top_choice_point)
#define LOCAL_top_or_fr                  (LOCAL->top_or_frame)
#define LOCAL_load                       (LOCAL->load)
#define LOCAL_request                    (LOCAL->request_from_worker)
#define LOCAL_reply_signal               (LOCAL->sharing_signals.reply)
#define LOCAL_p_fase_signal              (LOCAL->sharing_signals.P_fase)
#define LOCAL_q_fase_signal              (LOCAL->sharing_signals.Q_fase)
#define LOCAL_lock_signals               (LOCAL->sharing_signals.lock)
#define LOCAL_start_global_copy          (LOCAL->global_copy.start)
#define LOCAL_end_global_copy            (LOCAL->global_copy.end)
#define LOCAL_start_local_copy           (LOCAL->local_copy.start)
#define LOCAL_end_local_copy             (LOCAL->local_copy.end)
#define LOCAL_start_trail_copy           (LOCAL->trail_copy.start)
#define LOCAL_end_trail_copy             (LOCAL->trail_copy.end)
#define LOCAL_cut_cp                     (LOCAL->cut_choice_point)
#define LOCAL_top_cp_on_stack            (LOCAL->top_choice_point_on_stack)
#define LOCAL_top_collected_susp_fr      (LOCAL->top_collected_suspension_frame)
#define LOCAL_top_dep_fr                 (LOCAL->top_dependency_frame)
#define LOCAL_top_sg_fr                  (LOCAL->top_subgoal_frame)

#define REMOTE_lock(i)                   (REMOTE[i]->lock)
#define REMOTE_top_cp(i)                 (REMOTE[i]->top_choice_point)
#define REMOTE_top_or_fr(i)              (REMOTE[i]->top_or_frame)
#define REMOTE_load(i)                   (REMOTE[i]->load)
#define REMOTE_request(i)                (REMOTE[i]->request_from_worker)
#define REMOTE_reply_signal(i)           (REMOTE[i]->sharing_signals.reply)
#define REMOTE_p_fase_signal(i)          (REMOTE[i]->sharing_signals.P_fase)
#define REMOTE_q_fase_signal(i)          (REMOTE[i]->sharing_signals.Q_fase)
#define REMOTE_lock_signals(i)           (REMOTE[i]->sharing_signals.lock)
#define REMOTE_start_global_copy(i)      (REMOTE[i]->global_copy.start)
#define REMOTE_end_global_copy(i)        (REMOTE[i]->global_copy.end)
#define REMOTE_start_local_copy(i)       (REMOTE[i]->local_copy.start)
#define REMOTE_end_local_copy(i)         (REMOTE[i]->local_copy.end)
#define REMOTE_start_trail_copy(i)       (REMOTE[i]->trail_copy.start)
#define REMOTE_end_trail_copy(i)         (REMOTE[i]->trail_copy.end)
#define REMOTE_cut_cp(i)                 (REMOTE[i]->cut_choice_point)
#define REMOTE_top_cp_on_stack(i)        (REMOTE[i]->top_choice_point_on_stack)
#define REMOTE_top_collected_susp_fr(i)  (REMOTE[i]->top_collected_suspension_frame)
#define REMOTE_top_dep_fr(i)             (REMOTE[i]->top_dependency_frame)
#define REMOTE_top_sg_fr(i)              (REMOTE[i]->top_subgoal_frame)



#ifdef YAPOR
#include "or.structs.h"
#endif /* YAPOR */



#ifdef TABLING
#include "tab.structs.h"
#endif  /* TABLING */
