/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		load_dl.c						 *
* comments:	dl based dynamic loader  of external routines		 *
*               tested on i486-linuxelf					 *
*************************************************************************/

#include "Yap.h"
#include "Foreign.h"

#if LOAD_DLL

#include <windows.h>
#ifndef FILENAME_MAX
#define FILENAME_MAX 1024
#endif

/*
 *   YAPFindExecutable(argv[0]) should be called on yap initialization to
 *   locate the executable of Yap
*/
void
YAPFindExecutable(char *name)
{
  /* not really needed for dl version */
  strcpy(YapExecutable,".\\yap");
}


/*
 * LoadForeign(ofiles,libs,proc_name,init_proc) dynamically loads foreign
 * code files and libraries and locates an initialization routine
*/
Int
LoadForeign(StringList ofiles, StringList libs,
	       char *proc_name,	YapInitProc *init_proc)
{

  while (ofiles) {
    HINSTANCE handle;
    char fullpath[FILENAME_MAX];

    /* dlopen wants to follow the LD_CONFIG_PATH */
    if (!TrueFileName(ofiles->s, fullpath)) {
      strcpy(LoadMsg, "[ Trying to open unexisting file in LoadForeign ]");
      return LOAD_FAILLED;
    }
    if((handle=LoadLibrary(fullpath)) == 0)
    {
/*      strcpy(LoadMsg,dlerror());*/
      return LOAD_FAILLED;
    }

    if (*init_proc != NULL)
      *init_proc = (YapInitProc)GetProcAddress((HMODULE)handle, proc_name);

    ofiles = ofiles->next;
  }

  /* load libraries first so that their symbols are available to
     other routines */
  while (libs) {
    HINSTANCE handle;
    char s[FILENAME_MAX];

    if (libs->s[0] == '-') {
      strcat(s,libs->s+2);
      strcat(s,".dll");
    } else {
      strcpy(s,libs->s);
    }

    if((handle=LoadLibrary(s)) == 0)
    {
/*      strcpy(LoadMsg,dlerror());*/
      return LOAD_FAILLED;
    }

    if (*init_proc != NULL)
      *init_proc = (YapInitProc)GetProcAddress((HMODULE)handle, proc_name);

    libs = libs->next;
  }

  if(*init_proc == NULL) {
    strcpy(LoadMsg,"Could not locate initialization routine");
    return LOAD_FAILLED;
  }

  return LOAD_SUCCEEDED;
}

#endif

