/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		utilpreds.c						 *
* Last rev:	4/03/88							 *
* mods:									 *
* comments:	new utility predicates for YAP				 *
*									 *
*************************************************************************/
#ifdef SCCS
static char     SccsId[] = "@(#)utilpreds.c	1.3";
#endif

#include "Yap.h"
#include "Yatom.h"
#include "Heap.h"
#include "yapio.h"
#include "eval.h"
#if COROUTINING
#include "corout_utils.h"
#endif

typedef struct {
	Term            old_var;
	Term            new_var;
}              *vcell;


STATIC_PROTO(void  copy_complex_term, (CELL *, CELL *, CELL *, CELL *));
STATIC_PROTO(CELL  vars_in_complex_term, (CELL *, CELL *));
STATIC_PROTO(Int   p_non_singletons_in_term, (void));
STATIC_PROTO(CELL  non_singletons_in_complex_term, (CELL *, CELL *));
STATIC_PROTO(Int   p_variables_in_term, (void));
STATIC_PROTO(Int   ground_complex_term, (CELL *, CELL *));
STATIC_PROTO(Int   p_ground, (void));
STATIC_PROTO(Int   p_copy_term, (void));
STATIC_PROTO(Int   var_in_complex_term, (CELL *, CELL *, Term));

#ifdef DEBUG
STATIC_PROTO(Int  p_force_trail_expansion, (void));
#endif /* DEBUG */

#ifdef RATIONAL_TREES

#define UNWIND_CUNIF()                                        \
         while (visited < AuxSp) {                            \
            pt0 = (CELL *)visited[0];                         \
            *pt0 = visited[1];                                \
            visited += 2;                                     \
         }

#else
#define UNWIND_CUNIF()
#endif

static void copy_complex_term(register CELL *pt0, register CELL *pt0_end, CELL *ptf, CELL *HLow)
{

  CELL **to_visit = (CELL **)(HeapTop + sizeof(CELL));
  tr_fr_ptr TR0 = TR, TRB;
#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif
  CELL *HB0 = HB;
  HB = HLow;
  TR += 1024;
  TRB = TR;

 loop:
  while (pt0 < pt0_end) {
    register CELL d0;
    register CELL *ptd0;
    ++ pt0;
    ptd0 = pt0;
    d0 = *ptd0;
#ifdef COROUTINING
  continue_loop:
#endif
    deref_head(d0, copy_term_unk);
  copy_term_nvar:
    {
      if (IsPairTerm(d0)) {
	CELL *ap2 = RepPair(d0);
	if (ap2 >= HB && ap2 < H) {
	  /* If this is newer than the current term, just reuse */
	  *ptf++ = d0;
	  continue;
	} 
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = AbsPair(H);
	
#endif
	*ptf = AbsPair(H);
	ptf++;
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit[2] = ptf;
	  to_visit += 3;
	}
	pt0 = ap2 - 1;
	pt0_end = ap2 + 1;
	ptf = H;
	H += 2;
      } else if (IsApplTerm(d0)) {
	register Functor f;
	register CELL *ap2;
	/* store the terms to visit */
	ap2 = RepAppl(d0);
	if (ap2 >= HB && ap2 < H) {
	  /* If this is newer than the current term, just reuse */
	  *ptf++ = d0;
	  continue;
	} 
	f = (Functor)(*ap2);

	if (IsExtensionFunctor(f)) {
#if COROUTINING
	  if (IsAttachFunc(f)) {
	    CELL *pt2;
	    pt2 = attas[ExtFromFunctor(f)].sderef_op(d0);
	    if (IsVarTerm(SDerefa(pt2))) {
	      /* if unbound, call the standard copy term routine */
	      CELL **bp[1];
	      tr_fr_ptr CurTR = TR;
	      HB = HB0;
	      TR = TR0;

	      bp[0] = to_visit;
	      *ptf++ = attas[ExtFromFunctor(f)].copy_term_op(d0, bp);
	      to_visit = bp[0];
	      HB = HLow;
	      TR0 = TR;
	      /*	      if (TR > TRB) GapShouldHaveBeenLarger(); */
	      TR = CurTR;
	      Bind(pt2,ptf[-1]);
	      /* this routine should also old bound to new */
	    } else {
	      /* if bound ignore it */
	      d0 = SDerefa(pt2);
	      goto continue_loop;
	    }
	  } else
#endif
	    {
	      *ptf++ = d0;  /* you can just copy other extensions. */
	    }
	  continue;
	}
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = AbsAppl(H);
#endif
	*ptf = AbsAppl(H);
	ptf++;
	/* store the terms to visit */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit[2] = ptf;
	  to_visit += 3;
	}
	d0 = ArityOfFunctor(f);
	pt0 = ap2;
	pt0_end = ap2 + d0;
	/* store the functor for the new term */
	H[0] = (CELL)f;
	ptf = H+1;
	H += 1+d0;
      } else {
	/* just copy atoms or integers */
	*ptf++ = d0;
      }
      continue;
    }
    

    derefa_body(d0, ptd0, copy_term_unk, copy_term_nvar);
    if (ptd0 >= HLow && ptd0 < H) { 
      /* we have already found this cell */
      *ptf++ = (CELL) ptd0;
    } else {
      /* first time we met this term */
      RESET_VARIABLE(ptf);
      Bind(ptd0, (CELL)ptf);
      ptf++;
    }
  }
  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)(HeapTop + sizeof(CELL))) {
    to_visit -= 3;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    ptf = to_visit[2];
    goto loop;
  }

  UNWIND_CUNIF();
  /* we've done it */
  /* restore our nice, friendly, term to its original state */
  HB = HB0;
  if (TR != TRB) {
    do {
      CELL *p = (CELL *)TrailTerm(--TR);
      RESET_VARIABLE(p);
    } while (TR != TRB);
  }
  TR = TR0;
}

static Term
CopyTerm(Term inp) {
#if COROUTINING
  Term t = SDeref(inp);
#else
  Term t = Deref(inp);
#endif

#if TERM_EXTENSIONS
 restart:
#endif
  if (IsVarTerm(t)) {
    return(MkVarTerm());
  } else if (IsPrimitiveTerm(t)) {
    return(t);
  } else if (IsPairTerm(t)) {
    Term tf;
    CELL *ap = RepPair(t);
    tf = AbsPair(H);
    H += 2;
    copy_complex_term(ap-1, ap+1, H-2, H-2);
    return(tf);
  } else {
    Functor f = FunctorOfTerm(t);
    Term tf;
    CELL *HB0 = H;
    CELL *ap = RepAppl(t);
    tf = AbsAppl(H);
#if TERM_EXTENSIONS
    if (IsAttachFunc(f)) {
      CELL *pt2;
      pt2 = attas[ExtFromFunctor(f)].sderef_op(t);
      if (IsVarTerm(SDerefa(pt2))) {
	Term t0[2];
	t0[0] = t;
	copy_complex_term(t0-1, t0, t0+1, HB0);
	return(t0[1]);
      } else {
	t = SDerefa(pt2);
	goto restart;
      }
    }
#endif
    H[0] = (CELL)f;
    H += 1+ArityOfFunctor(f);
    copy_complex_term(ap, ap+ArityOfFunctor(f), HB0+1, HB0);
    return(tf);
  }
}
 
static Int 
p_copy_term(void)		/* copy term t to a new instance  */
{
  return(unify(ARG2,CopyTerm(ARG1)));
}


static Term vars_in_complex_term(register CELL *pt0, register CELL *pt0_end)
{

  register CELL **to_visit = (CELL **)(HeapTop + sizeof(CELL));
  register tr_fr_ptr TR0 = TR;
  CELL *InitialH = H;
  CELL output = AbsPair(H);
#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif

 loop:
  while (pt0 < pt0_end) {
    register CELL d0;
    register CELL *ptd0;
    ++ pt0;
    ptd0 = pt0;
    d0 = *ptd0;
#ifdef COROUTINING
  continue_loop:
#endif
    deref_head(d0, vars_in_term_unk);
  vars_in_term_nvar:
    {
      if (IsPairTerm(d0)) {
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
	
#endif
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	pt0 = RepPair(d0) - 1;
	pt0_end = RepPair(d0) + 1;
      } else if (IsApplTerm(d0)) {
	register Functor f;
	register CELL *ap2;
	/* store the terms to visit */
	ap2 = RepAppl(d0);
	f = (Functor)(*ap2);

	if (IsExtensionFunctor(f)) {
#if COROUTINING
	  if (IsAttachFunc(f)) {
	    CELL *pt2;
	    pt2 = attas[ExtFromFunctor(f)].sderef_op(d0);
	    if (IsVarTerm(SDerefa(pt2))) {
	      H[1] = AbsPair(H+2);
	      H += 2;
	      H[-2] = d0;
	      *pt2 = TermNil;
	      TrailTerm(TR++) = (CELL)pt2;
	    } else {
	      d0 = SDerefa(pt2);
	      goto continue_loop;
	    }
	  }
#endif

	  continue;
	}
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	/* store the terms to visit */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	d0 = ArityOfFunctor(f);
	pt0 = ap2;
	pt0_end = ap2 + d0;
      }
      continue;
    }
    

    derefa_body(d0, ptd0, vars_in_term_unk, vars_in_term_nvar);
    /* do or pt2 are unbound  */
    *ptd0 = TermNil;
    /* leave an empty slot to fill in later */
    H[1] = AbsPair(H+2);
    H += 2;
    H[-2] = (CELL)ptd0;
    /* next make sure noone will see this as a variable again */ 
    TrailTerm(TR++) = (CELL)ptd0;
  }
  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)(HeapTop + sizeof(CELL))) {
    to_visit -= 2;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    goto loop;
  }

  UNWIND_CUNIF();
  /* we've done it */
  /* restore our nice, friendly, term to its original state */
  if (TR != TR0) {
    do {
      CELL *p = (CELL *)TrailTerm(--TR);
      RESET_VARIABLE(p);
    } while (TR != TR0);
  }
  if (H != InitialH) {
    /* close the list */
    Term t2 = Deref(ARG2);
    if (IsVarTerm(t2)) {
      RESET_VARIABLE(H-1);
      unify((CELL)(H-1),ARG2);
    } else {
      H[-1] = t2;		/* don't need to trail */
    }
    return(output);
  } else {
    return(ARG2);
  }
}
 
static Int 
p_variables_in_term(void)	/* variables in term t		 */
{
#if COROUTINING
  Term t = SDeref(ARG1);
#else
  Term t = Deref(ARG1);
#endif
  Term out;

#if TERM_EXTENSIONS
 restart:
#endif
  if (IsVarTerm(t)) {
    out = AbsPair(H);
    H += 2;
    RESET_VARIABLE(H-2);
    RESET_VARIABLE(H-1);
    unify((CELL)(H-2),ARG1);
    unify((CELL)(H-1),ARG2);
  }  else if (IsPrimitiveTerm(t)) 
    out = ARG2;
  else if (IsPairTerm(t)) {
    out = vars_in_complex_term(RepPair(t)-1,
			       RepPair(t)+1);
  }
  else {
    Functor f = FunctorOfTerm(t);
#if TERM_EXTENSIONS
    if (IsAttachFunc(f)) {
      t = SDerefa(attas[ExtFromFunctor(f)].sderef_op(t));
      goto restart;
    }
#endif
    out = vars_in_complex_term(RepAppl(t),
			       RepAppl(t)+
			       ArityOfFunctor(f));
  }
  return(unify(ARG3,out));
}


static Term non_singletons_in_complex_term(register CELL *pt0, register CELL *pt0_end)
{

  register CELL **to_visit = (CELL **)(HeapTop + sizeof(CELL));
  register tr_fr_ptr TR0 = TR;
  CELL *InitialH = H;
  CELL output = AbsPair(H);
#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif

 loop:
  while (pt0 < pt0_end) {
    register CELL d0;
    register CELL *ptd0;
    ++ pt0;
    ptd0 = pt0;
    d0 = *ptd0;
#ifdef COROUTINING
  continue_loop:
#endif
    deref_head(d0, vars_in_term_unk);
  vars_in_term_nvar:
    {
      if (IsPairTerm(d0)) {
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	pt0 = RepPair(d0) - 1;
	pt0_end = RepPair(d0) + 1;
      } else if (IsApplTerm(d0)) {
	register Functor f;
	register CELL *ap2;
	/* store the terms to visit */
	ap2 = RepAppl(d0);
	f = (Functor)(*ap2);

	if (IsExtensionFunctor(f)) {
#ifdef COROUTINING
	  if (IsAttachFunc(f)) {
	    CELL *pt2;
	    pt2 = attas[ExtFromFunctor(f)].sderef_op(d0);
	    if (*pt2 == TermFoundVar) {
	      H[1] = AbsPair(H+2);
	      H += 2;
	      H[-2] = (CELL)pt2;
	      *pt2 = TermReFoundVar;
	    } else {
	      d0 = SDerefa(pt2);
	      goto continue_loop;
	    }
	  }
#endif

	  continue;
	}
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	/* store the terms to visit */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	d0 = ArityOfFunctor(f);
	pt0 = ap2;
	pt0_end = ap2 + d0;
      } else if (d0 == TermFoundVar) {
	CELL *pt2 = pt0;
	while(IsVarTerm(*pt2))
	  pt2 = (CELL *)(*pt2);
	H[1] = AbsPair(H+2);
	H += 2;
	H[-2] = (CELL)pt2;
	*pt2 = TermReFoundVar;
      }
      continue;
    }
    

    derefa_body(d0, ptd0, vars_in_term_unk, vars_in_term_nvar);
    /* do or pt2 are unbound  */
    *ptd0 = TermFoundVar;
    /* next make sure we can recover the variable again */ 
    TrailTerm(TR++) = (CELL)ptd0;
  }
  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)(HeapTop + sizeof(CELL))) {
    to_visit -= 2;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    goto loop;
  }

  UNWIND_CUNIF();
  /* we've done it */
  /* restore our nice, friendly, term to its original state */
  if (TR != TR0) {
    do {
      CELL *p = (CELL *)TrailTerm(--TR);
      RESET_VARIABLE(p);
    } while (TR != TR0);
  }
  if (H != InitialH) {
    /* close the list */
    RESET_VARIABLE(H-1);
    unify((CELL)(H-1),ARG2);
    return(output);
  } else {
    return(ARG2);
  }
}
 
static Int 
p_non_singletons_in_term(void)	/* non_singletons in term t		 */
{
        Term t = Deref(ARG1);
	Term out;
	if (IsVarTerm(t)) {
	  out = MkPairTerm(t,ARG2);
	}  else if (IsPrimitiveTerm(t)) 
	  out = ARG2;
	else if (IsPairTerm(t)) {
	  out = non_singletons_in_complex_term(RepPair(t)-1,
				     RepPair(t)+1);
	}
	else out = non_singletons_in_complex_term(RepAppl(t),
					RepAppl(t)+
					ArityOfFunctor(FunctorOfTerm(t)));
	return(unify(ARG3,out));
}

static Int ground_complex_term(register CELL *pt0, register CELL *pt0_end)
{

  register CELL **to_visit = (CELL **)(HeapTop + sizeof(CELL));
#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif

 loop:
  while (pt0 < pt0_end) {
    register CELL d0;
    register CELL *ptd0;
    ++ pt0;
    ptd0 = pt0;
    d0 = *ptd0;
#ifdef COROUTINING
  continue_loop:
#endif
    deref_head(d0, vars_in_term_unk);
  vars_in_term_nvar:
    {
      if (IsPairTerm(d0)) {
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	pt0 = RepPair(d0) - 1;
	pt0_end = RepPair(d0) + 1;
      } else if (IsApplTerm(d0)) {
	register Functor f;
	register CELL *ap2;
	/* store the terms to visit */
	ap2 = RepAppl(d0);
	f = (Functor)(*ap2);

	if (IsExtensionFunctor(f)) {
#ifdef COROUTINING
	  if (IsAttachFunc(f)) {
	  d0 = Deref(d0);
	  goto continue_loop;
	}
#endif
	  continue;
	}
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	/* store the terms to visit */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	d0 = ArityOfFunctor(f);
	pt0 = ap2;
	pt0_end = ap2 + d0;
      }
      continue;
    }
    

    derefa_body(d0, ptd0, vars_in_term_unk, vars_in_term_nvar);
    UNWIND_CUNIF();
    return(FALSE);
  }
  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)(HeapTop + sizeof(CELL))) {
    to_visit -= 2;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    goto loop;
  }
  UNWIND_CUNIF();
  return(TRUE);
}
 
static Int 
p_ground(void)			/* ground(+T)		 */
{
        Term t = Deref(ARG1);

	if (IsVarTerm(t)) {
	  return(FALSE);
	}  else if (IsPrimitiveTerm(t)) {
	  return(TRUE);
	} else if (IsPairTerm(t)) {
	  return(ground_complex_term(RepPair(t)-1,
				     RepPair(t)+1));
	} else {
	  Functor fun = FunctorOfTerm(t);
	  if (IsExtensionFunctor(fun))
	    return(TRUE);
	  else return(ground_complex_term(RepAppl(t),
					  RepAppl(t)+
					  ArityOfFunctor(fun)));
	}
}

static Int var_in_complex_term(register CELL *pt0, register CELL *pt0_end, Term t)
{

  register CELL **to_visit = (CELL **)(HeapTop + sizeof(CELL));
#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif

 loop:
  while (pt0 < pt0_end) {
    register CELL d0;
    register CELL *ptd0;
    ++ pt0;
    ptd0 = pt0;
    d0 = *ptd0;
#ifdef COROUTINING
  continue_loop:
#endif
    deref_head(d0, vars_in_term_unk);
  vars_in_term_nvar:
    {
      if (IsPairTerm(d0)) {
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	pt0 = RepPair(d0) - 1;
	pt0_end = RepPair(d0) + 1;
      } else if (IsApplTerm(d0)) {
	register Functor f;
	register CELL *ap2;
	/* store the terms to visit */
	ap2 = RepAppl(d0);
	f = (Functor)(*ap2);

	if (IsExtensionFunctor(f)) {
#ifdef COROUTINING
	  if (f == FunctorDelay || f == FunctorAttVar
#ifdef YAP_ARRAYS
	      || f == FunctorArrayAccess
#endif
	      ) {
	    d0 = Deref(d0);
	    goto continue_loop;
	  }
#endif

	  continue;
	}
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	/* store the terms to visit */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	d0 = ArityOfFunctor(f);
	pt0 = ap2;
	pt0_end = ap2 + d0;
      }
      continue;
    }
    

    deref_body(d0, ptd0, vars_in_term_unk, vars_in_term_nvar);
    if ((CELL)ptd0 == t) { /* we found it */
      UNWIND_CUNIF();
      return(TRUE);
    }
  }
  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)(HeapTop + sizeof(CELL))) {
    to_visit -= 2;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    goto loop;
  }
  UNWIND_CUNIF();
  return(FALSE);
}
 
Int 
var_in_term(Term v, Term t)		/* variables in term t		 */
{

  if (IsPrimitiveTerm(t)) 
    return(FALSE);
  else if (IsPairTerm(t)) {
    return(var_in_complex_term(RepPair(t)-1,
			       RepPair(t)+1, v));
  }
  else return(var_in_complex_term(RepAppl(t),
				  RepAppl(t)+
				  ArityOfFunctor(FunctorOfTerm(t)),v));
}

/* The code for TermHash was originally contributed by Gertjen Van Noor */

/* This code with max_depth == -1 will loop for infinite trees */

#define GvNht ((UInt *)H)

#define HASHADD(T) (GvNht[k]+=(T), k=(k<2 ? k+1 : 0))

static Int TermHash(Term t1, Int depth_lim, Int k)
{
  Int i;
  if (IsVarTerm(t1)) {
    return(-1);
  } else if (IsAtomTerm(t1)) {
    register char *s = AtomName(AtomOfTerm(t1));
    for (i=0; s[i]; i++)
	HASHADD(s[i]);
    return(k);
  } else if (IsPairTerm(t1)) {
    HASHADD('.');
    depth_lim--;
    if (depth_lim == 0) return(TRUE);
    k = TermHash(HeadOfTerm(t1),depth_lim,k);
    if (k < 0) return(-1);
    k = TermHash(TailOfTerm(t1),depth_lim,k);
    return(k);
  } else if (IsIntTerm(t1)) {
    HASHADD(IntOfTerm(t1));
    return(k);
  } else {
    Functor f = FunctorOfTerm(t1);

    if (IsExtensionFunctor(f)) {
      if (f == FunctorDouble) {
	Int *iptr = (Int *)(RepAppl(t1)+1);
	int i;

	for (i = 0; i < sizeof(Float) / sizeof(CELL); i++) {
	  HASHADD(*iptr++);
	}
	return(k);
      } else if (f == FunctorLongInt) {
	HASHADD(LongIntOfTerm(t1));
	return(k);
      } else if (f == FunctorDBRef) {
	HASHADD((Int)DBRefOfTerm(t1));
	return(k);
	/* should never happen */
      } else {
	return(-1);
      }
    } else {
      int ar = ArityOfFunctor(FunctorOfTerm(t1));
      int res = TRUE;
      register char *s = AtomName(NameOfFunctor(f));

      depth_lim--;
      if (depth_lim == 0) return(TRUE);
      for (i=0; s[i]; i++) 
	HASHADD(s[i]);
      for (i=1; i<=ar && res; i++) {
	k = TermHash(ArgOfTerm(i,t1),depth_lim,k);
	if (k == -1 ) return(-1);
      }
      return(k);
    }
  }
}

static Int
GvNTermHash(void)
{
  unsigned int i1,i2,i3;
  Term t1 = Deref(ARG1);
  Term t2 = Deref(ARG2);
  Term t3 = Deref(ARG3);
  Term result;
  Int size, depth;


  if (IsVarTerm(t2)) {
    Error(INSTANTIATION_ERROR,t2,"term_hash/4");
    return(FALSE);
  }
  if (!IsIntTerm(t2)) {
    Error(TYPE_ERROR_INTEGER,t2,"term_hash/4");
    return(FALSE);
  }
  depth = IntOfTerm(t2);
  if (depth == 0) {
    if (IsVarTerm(t1)) return(TRUE);
    return(unify(ARG4,MkIntTerm(0)));
  }
  if (IsVarTerm(t3)) {
    Error(INSTANTIATION_ERROR,t3,"term_hash/4");
    return(FALSE);
  }
  if (!IsIntTerm(t3)) {
    Error(TYPE_ERROR_INTEGER,t3,"term_hash/4");
    return(FALSE);
  }
  size = IntOfTerm(t3);
  GvNht[0] = 0;
  GvNht[1] = 0;
  GvNht[2] = 0;

  if (TermHash(t1,depth,0) == -1) return(TRUE);

  i1 = GvNht[0];
  i2 = GvNht[1];
  i3 = GvNht[2];
  i2 ^= i3; i1 ^= i2; i1 = (((i3 << 7) + i2) << 7) + i1;
  result = MkIntTerm(i1 % size);
  return(unify(ARG4,result));
}


#ifdef DEBUG
static Int
p_force_trail_expansion()
{
  Int i = IntOfTerm(Deref(ARG1))*1024, j = 0;
  tr_fr_ptr OTR = TR;

  for (j = 0; j < i; j++) {
    TrailTerm(TR) = 0;
    TR++;
  }
  TR = OTR;

  return(TRUE);
}
#endif /* DEBUG */

void InitUtilCPreds(void)
{
  InitCPred("copy_term", 2, p_copy_term, SafePredFlag);
  InitCPred("ground", 1, p_ground, SafePredFlag);
  InitCPred("$variables_in_term", 3, p_variables_in_term, SafePredFlag);
  InitCPred("$non_singletons_in_term", 3, p_non_singletons_in_term, SafePredFlag);
  InitCPred("term_hash", 4, GvNTermHash, SafePredFlag);
#ifdef DEBUG
  InitCPred("$force_trail_expansion", 1, p_force_trail_expansion, SafePredFlag);
#endif
}

