/*************************************************************************
*									 *
*	 Yap Prolog 							 *
*									 *
*	Yap Prolog Was Developed At Nccup - Universidade Do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa And Universidade Do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		Yap.C							 *
* Last Rev:								 *
* Mods:									 *
* Comments:	Yap's Main File						 *
*									 *
*************************************************************************/
/* static char SccsId[] = "X 4.1.19"; */

#include "Yap.h"
#include "Yatom.h"
#include "Heap.h"
#include "Foreign.h"
#include "yapio.h"
#if defined(YAPOR) || defined(TABLING)
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#endif	/* YAPOR || TABLING */

#if HAVE_STDARG_H
#include <stdarg.h>
#endif
#include <stdlib.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

STATIC_PROTO(int mygetc, (int));
STATIC_PROTO(void do_bootfile, (char *));
STATIC_PROTO(void do_top_goal,(Term));
STATIC_PROTO(void enter_yap,(int, char *));

#ifndef LIGHT
STD_PROTO (void exit, (int));
#endif

#ifdef LIGHT
#include <unix.h>
int
  _main (int, char **);
#else
STD_PROTO (int main, (int, char **));
#endif

char BootFile[] = "boot.yap";
char StartUpFile[] = "startup";

#ifdef lint
/* VARARGS1 */
#endif

static int pos = 0, eof_found = 0;

#ifdef M_WILLIAMS
long _stksize = 32000;
#endif

static FILE *bootfile;

static int
mygetc (int sno)
{
  int ch;
  if (eof_found)
    return (-1);
  ch = YP_getc (bootfile);
  if (ch == -1)
    eof_found = TRUE;
#ifdef MPW
  if (ch == '\n')
    ch = 10;
#endif
  return (ch);
}

static void
do_top_goal (Term Goal)
{
#ifdef DEBUG
  if (output_msg)
    YP_fprintf(YP_stderr,"Entering absmi\n");
#endif
  if (TrailTop - HeapTop < 2048)
    {
      PrologMode = BootMode;
      Error(SYSTEM_ERROR,TermNil,
	    "unable to boot because of too little heap space");
    }
  /* PlPutc(0,'a'); PlPutc(0,'\n'); */
  YapRunTopGoal(Goal);
}

/* do initial boot by consulting the file boot.yap */
static void
do_bootfile (char *bootfilename)
{
  Term t;
  TokEntry *tokstart;

  /* set yaam registers for consulting */
  InitYaamRegs();
  /* consult boot.pl */
  bootfile = YP_fopen (bootfilename, "r");
  if (bootfile == NIL)
    {
#if HAVE_STRERROR
      Error(SYSTEM_ERROR, TermNil, "unable to boot using %s, %s", bootfilename, strerror(errno));
#else
      Error(SYSTEM_ERROR, TermNil, "unable to boot using %s, %s", strerror(errno));
#endif
      exit(1);
    }
  PutValue (AtomIndex, MkAtomTerm (AtomTrue));
  while (!eof_found)
    {
      pos = 0;
      ParserAuxSp = ASP - 64;
      tokptr = toktide = tokstart = tokenizer (FunAdr(mygetc), FunAdr(mygetc));
      if (eof_found)
	break;
      if (ErrorMessage)
	{
	  YP_fprintf(YP_stderr, "%s\n", ErrorMessage);

	  break;
	}
      t = Parse ();
      if (t == 0 || IsVarTerm (t))
	{
	  YP_fprintf(YP_stderr, "** syntax error**\n");
	  syntax_error (tokstart);
	}
      else if (t == MkAtomTerm (AtomTrue))
	{
	  exit_yap (0, NIL);
	}
      else if (IsPairTerm (t))
	Error(FATAL_ERROR, TermNil, "consult not allowed in boot.pl");
      else if (IsApplTerm (t) && FunctorOfTerm (t) == FunctorQuery) 
	{ 
	  do_top_goal(ArgOfTerm (1, t));
        }
      else
	{
	  CODEADDR codeaddr;
	  
	  ARG1 = t;
	  codeaddr = cclause (t,0);
	  t = Deref(ARG1); /* just in case there was an heap overflow */
	  if (ErrorMessage)
	    YP_fprintf(YP_stderr, ErrorMessage);
	  else
	    addclause (t, codeaddr, TRUE);
	  if (ErrorMessage)
	    YP_fprintf(YP_stderr, ErrorMessage);
	}
    }
  YP_fclose (bootfile);
#ifdef DEBUG
  if (output_msg)
    YP_fprintf(YP_stderr,"Boot loaded\n");
#endif
}

int in_linbo = 1;

static int Heap = 0, Stack = 0, Trail = 0;
static char *filename;
#if defined(YAPOR) || defined(TABLING)
static int  aux_number_workers = 1;
static int  aux_scheduler_loop = 10;
static int  aux_delayed_release_load = 3;
#endif /* YAPOR || TABLING */

/*
 * proccess command line arguments: valid switches are: -b    boot -s
 * stack area size (K) -h    heap area size -a    aux stack size -e
 * emacs_mode -m    reserved memory for alloc IF DEBUG -p    if you
 * want to check out startup IF MAC -mpw  if we are using the mpw
 * shell
 */
static int
ParseYapArguments(int argc, char *argv[])
{
  char *p;
  int PrologBoot = FALSE;
  int *ssize;

  yap_args = argv;
  yap_argc = argc;
  while (--argc > 0)
    {
      p = *++argv;
      if (*p == '-')
	switch (*++p)
	  {
	  case 'b':
	    PrologBoot = TRUE;
	    break;
          case '?':
            YP_fprintf(YP_stderr,"\n[ Valid switches for command line arguments: ]\n");
            YP_fprintf(YP_stderr,"  -?   Shows this screen\n");
            YP_fprintf(YP_stderr,"  -b   Boot file (%s)\n", StartUpFile);
            YP_fprintf(YP_stderr,"  -h   Heap area in Kbytes (%d)\n", DefHeapSpace);
	    YP_fprintf(YP_stderr,"  -s   Stack area in Kbytes (%d)\n", DefStackSpace);
	    YP_fprintf(YP_stderr,"  -t   Aux/trail area in Kbytes (%d)\n", DefTrailSpace);
#ifdef YAPOR
            YP_fprintf(YP_stderr,"  -w   Number of workers (%d)\n", aux_number_workers);
            YP_fprintf(YP_stderr,"  -l   Loop scheduler executions before look for hiden shared work (%d)\n", aux_scheduler_loop);
            YP_fprintf(YP_stderr,"  -d   Value of delayed release of load (%d)\n", aux_delayed_release_load);
#endif /* YAPOR */
            YP_fprintf(YP_stderr,"\n");
            exit(0);
#ifdef YAPOR
	  case 'w':
	    ssize = &aux_number_workers;
	    goto GetSize;
          case 'l':
            ssize = &aux_scheduler_loop;  
	    goto GetSize;
          case 'd':
            ssize = &aux_delayed_release_load;
	    goto GetSize;
#endif /* YAPOR */
#ifdef HAVE_SOCKET
          case 'c':          /* running as client */
	    {
	      char *host, *p1;
	      long port;
	      char *ptr;

	      host = *++argv;
	      argc--;
	      if (host != NULL && host[0] == '-')
		Error(FATAL_ERROR,TermNil,
		      "sockets must receive host to connect to");
	      p1 = *++argv;
	      argc--;
	      if (p1[0] == '-')
		Error(FATAL_ERROR,TermNil,
		      "sockets must receive port to connect to");
	      port = strtol(p1, &ptr, 10);
	      if (ptr == NULL || ptr[0] != '\0')
		Error(FATAL_ERROR,TermNil,
		      "port argument to socket must be a number");
	      init_socks(host,port);
	    }
	    break;
#endif
#ifdef EMACS
	  case 'e':
	    emacs_mode = TRUE;
	    {
	      YP_File fd;
	      strcpy (emacs_tmp, ++p);
	      if ((fd = YP_fopen (emacs_tmp, "w")) == NIL)
		YP_fprintf(YP_stderr, "[ Warning: unable to communicate with emacs: failed to open %s ]\n", emacs_tmp);
	      YP_fclose (fd);
	      unlink (emacs_tmp);
	      p = *++argv;
	      --argc;
	      strcpy (emacs_tmp2, p);
	      if ((fd = YP_fopen (emacs_tmp2, "w")) == NIL)
		YP_fprintf(YP_stderr, "Unable to communicate with emacs: failed to open %s\n", emacs_tmp2);
	      YP_fclose (fd);
	      unlink (emacs_tmp2);
	    }
	    break;
#endif /* EMACS */
#ifdef MPWSHELL
	  case 'm':
	    if (*++p == 'p' && *++p == 'w' && *++p == '\0')
	      mpwshell = TRUE;
	    break;
#endif
	  case 's':
	    ssize = &Stack;
	    goto GetSize;
	  case 'h':
	    ssize = &Heap;
	    goto GetSize;
	  case 't':
	    ssize = &Trail;
	  GetSize:
	    if (*++p == 0)
	      {
		if (argc > 1)
		  --argc, p = *++argv;
		else
		  {
		    Error(FATAL_ERROR, TermNil, "missing size in flag");
		    exit (1);
		  }
	      }
	    {
	      int i = 0, ch;
	      while ((ch = *p++) >= '0' && ch <= '9')
		i = i * 10 + ch - '0';
	      if (ch)
		{
		  Error(FATAL_ERROR, TermNil, "illegal size specification");
		  exit (1);
		}
	      *ssize = i;
	    }
	    break;
#ifdef DEBUG
	  case 'p':
	    output_msg = TRUE;
	    break;
#endif
	  case '-':
	    /* skip remaining arguments */
	    argc = 1;
	    break;
	  default:
	    Error(FATAL_ERROR, (Term)0, "unknown switch -%c", *p);
	    exit (1);
	  }
      else {
	filename = p;
      }
    }
  return(PrologBoot);
}

#if defined(YAPOR) || defined(TABLING)

static void
start_workers(void)
{
  make_root_frames();
#ifdef YAPOR
  init_workers();
#endif /* YAPOR */
  init_local();
  close_mapfile();
#ifdef YAPOR
  if (worker_id != 0) {
#if SBA
    /*
      In the SBA we cannot just happily inherit registers
      from the other workers
    */
    InitYaamRegs();
#endif
    /* slaves, waiting for work */
    CurrentModule = 1;
    in_linbo = 0;
    P = GETWORK_FIRST_TIME;
    exec_absmi(FALSE);
    abort_optyap("abstract machine unexpected exit");
  }
#endif /* YAPOR */
}
#endif /* YAPOR || TABLING */

static int
InitStandardSystem(int argc, char *argv[])
{
  int PrologBoot;

  PrologBoot = ParseYapArguments(argc,argv);

  /* init memory */
  if (PrologBoot)
    {
      if (!Trail)
	Trail = DefTrailSpace;
      if (!Stack)
	Stack = DefStackSpace;
      if (!Heap)
	Heap = DefHeapSpace;
    }
  else
    {
      if (!filename)
	filename = StartUpFile;
      if (SavedInfo (filename, &Trail, &Stack, &Heap) != 1) {
#if HAVE_STRERROR
	if (errno)
	  Error(FATAL_ERROR, TermNil, "could not open saved state %s, %s", filename, strerror(errno));
	else
#endif
	  Error(FATAL_ERROR, TermNil, "%s is not a saved state", filename);
	exit (1);
      }
    }

#if defined(YAPOR) || defined(TABLING)
    InitStacks (Heap, Stack, Trail, aux_number_workers, aux_scheduler_loop, aux_delayed_release_load);
#else
  InitStacks (Heap, Stack, Trail);
#endif

  return(PrologBoot);
}


static void
enter_yap(int PrologBoot, char *filename)
{
  CurrentModule = 1;
  in_linbo = 0;
  if (!PrologBoot && sigsetjmp(DeepRestart, 1) == 0) {
    /* load saved state */
    if (Restore(filename) == DO_EVERYTHING) {
      /* got the stacks ready, can enter absmi immediately */
      exec_absmi(TRUE);
    }
  }
  while (GetValue (AtomLive) != MkAtomTerm (AtomFalse))
    {
      do_top_goal (MkAtomTerm (AtomLive));
      if (PrologMode & BootMode)
	{
	  exit_yap (1, NIL);
	}
    }
  exit_yap (0, NIL);
}

#ifdef LIGHT
int
_main (int argc, char **argv)
#else
int
main (int argc, char **argv)
#endif
{
  int PrologBoot;

#ifdef SIMICS
  YP_fprintf(YP_stdout,"Entering YAP\n");
#endif /* SIMICS */
  PrologBoot = InitStandardSystem(argc, argv);
  if (PrologBoot)
    {
      do_bootfile (filename ? filename : BootFile);
      PutValue (AtomLive, MkAtomTerm (AtomTrue));
      PrologMode = UserMode;
    }
    
  InitYaamRegs();  
#if defined(YAPOR) || defined(TABLING)
  start_workers();
#endif /* YAPOR || TABLING */
  enter_yap(PrologBoot, filename);

  return(0);
}

